#include <iostream>
#include <string>
#include <vector>

//
// TrOp class contains an operation that may be
// undone during a rollback. It's an ABC from
// which all undoable operations derive.
//
class TrOp
{
  public:
    virtual void Undo() = 0;
    virtual ~TrOp() {}
};

//
// The Change template generates depending on
// the data type all possible operations (derived
// from TrOp) that may be undone. For example
// the instruction:
//
//           new Change<int>(a);
//
// generates a new "change of an integer" object
// and saves the old value to be able to restore
// it if an undo is requested.
//
template<class T> class Change : public TrOp
{
  public:
    T& who;
    T old_value;
    
    Change(T& s) : who(s), old_value(s) { }
    void Undo() { who=old_value; }
};

//
// The Transaction vector contains all the
// operations that must be undone to get
// back to the last good state of the
// database. It's a vector of pointers as
// TrOps are used polymorphically.
//
std::vector<TrOp *> Transaction;

//
// The Field template generates all possible
// fields in an database object. The Field<T>
// class is responsible for generating an
// appropriate Change<T> object before any
// modification.
//
template<class T> class Field : public T
{
  private:
    void Save()
    {
      Transaction.push_back(new Change<T>(*static_cast<T*>(this)));
    }
    Field& operator&();
  
  public:
    // Costructor from a T object
    Field(const T& s) : T(s) { }
    
    // Conversion to a T value (const!!!)
    operator const T() const { *this; }
    
    // Assigning
    Field& operator=(const T& v)
    { Save(); T::operator=(v); return *this; }
    
    Field& operator=(const Field<T>& v)
    { Save(); T::operator=(v); return *this; }
    
    // Remapping of ++/--
    const T operator++(int) { Save(); return T::operator++(0); }
    const T operator--(int) { Save(); return T::operator--(0); }
    Field& operator++() { Save(); T::operator++(); return *this; }
    Field& operator--() { Save(); T::operator--(); return *this; }
    
    // Remapping of modifiers
    Field& operator+=(const T& v) { Save(); T::operator+=(v); return *this; }
    Field& operator-=(const T& v) { Save(); T::operator-=(v); return *this; }
    Field& operator*=(const T& v) { Save(); T::operator*=(v); return *this; }
    Field& operator/=(const T& v) { Save(); T::operator/=(v); return *this; }
    Field& operator%=(const T& v) { Save(); T::operator%=(v); return *this; }

    //
    // Note: if there are methods that change
    // the object they should be overridden to
    // be able to call Save() before invoking
    // the original method of T.
    //
};

//
// Field specialization for basic types; this is needed
// because it's illegal to derive from a basic type.
//
template<class T> class BasicField
{
  protected:
    T v_;
    void Save()
    {
      Transaction.push_back(new Change<T>(v_));
    }
    BasicField& operator&();
  
  public:
    // Default constructor
    BasicField() : v_(T()) { }

    // Construct from a T object
    BasicField(const T& s) : v_(s) { }
    
    // Conversion to a T value (const!!!)
    operator const T() const { return v_; }
    
    // Assigning
    BasicField& operator=(const T& v)
    { Save(); v_=v; return *this; }
    
    BasicField& operator=(const BasicField<T>& v)
    { Save(); v_=v; return *this; }
    
    // Remapping ++/--
    const T operator++(int) { Save(); return v_++; }
    const T operator--(int) { Save(); return v_--; }
    BasicField& operator++() { Save(); ++v_; return *this; }
    BasicField& operator--() { Save(); --v_; return *this; }
    
    // Remapping modifiers
    BasicField& operator+=(const T& v) { Save(); v_+=v; return *this; }
    BasicField& operator-=(const T& v) { Save(); v_-=v; return *this; }
    BasicField& operator*=(const T& v) { Save(); v_*=v; return *this; }
    BasicField& operator/=(const T& v) { Save(); v_/=v; return *this; }
    BasicField& operator%=(const T& v) { Save(); v_%=v; return *this; }
    
    // Remapping output
    friend std::ostream& operator << (std::ostream& o,const BasicField& t)
    { return o << t.v_; }

};

//
// Template for pointers. This isn't a specialization because for
// some reason specializing Field<> doesn't work with MSVC++
//
template<class T> class Ptr : public BasicField<T*>
{
  public:
    typedef T *TPtr;
    Ptr(const TPtr& x) : BasicField<T*>(x) {  }
    
    Ptr& operator=(const TPtr& x)
    { BasicField<T*>::operator=(x); return *this; }
    
    Ptr& operator=(const BasicField<T*>& x)
    { BasicField<T*>::operator=(x); return *this; }
    
    // Rimappatura dereferencing
    T  operator*()  { return *v_; }
    T* operator->() { return v_; }
};

//
// Complete specialization for Field<int>. Other basic types
// require similar definition.
//
template<> class Field<int> : public BasicField<int>
{
  public:
    Field(const int& x) : BasicField<int>(x) {  }
    Field& operator=(const int& x)
    { BasicField<int>::operator=(x); return *this; }
    Field& operator=(const BasicField<int>& x)
    { BasicField<int>::operator=(x); return *this; }
};

void Rollback()
{
  while (!Transaction.empty())
  {
    Transaction.back()->Undo();
    delete Transaction.back();
    Transaction.pop_back();
  }
}

int main()
{
  Field<std::string> s("Initial value");
  Field<int> i(0);
  int j[4]={6,5,0,2};
  Ptr<int> p=j;
  
  p++;
  s+=" (modified)";
  i+=3;
  i++;

  int *pp=p;

  std::cout << " s = " << s << std::endl;
  std::cout << " i = " << i << std::endl;
  std::cout << " *p = " << *p << std::endl;
  std::cout << " *pp = " << *pp << std::endl;

  Rollback();
  
  std::cout << " s = " << s.c_str() << std::endl;
  std::cout << " i = " << i << std::endl;
  std::cout << " *p = " << *p << std::endl;
  std::cout << " *pp = " << *pp << std::endl;

  std::cout << "sizeof(Field<int>) = " << sizeof(Field<int>) << std::endl;
  return 0;
}
